import unittest

from avocado_varianter_cit.CombinationMatrix import CombinationMatrix
from avocado_varianter_cit.CombinationRow import CombinationRow


def combination_row_equals(row_1, row_2):
    return (row_1.covered_more_than_ones == row_2.covered_more_than_ones and row_1.uncovered == row_2.uncovered
            and row_1.hash_table == row_2.hash_table)


class MatrixInitialization(unittest.TestCase):

    def test_combination_matrix_initialization(self):
        """
        Test of proper initialization
        """
        data = [3, 3, 3, 4]
        t_value = 2
        matrix = CombinationMatrix(data, t_value)
        excepted_uncovered = 63
        excepted_covered_more_than_ones = 0
        excepted_row_size = 6
        excepted_hash_table = {(0, 1): CombinationRow(data, t_value, (0, 1)),
                               (0, 2): CombinationRow(data, t_value, (0, 2)),
                               (0, 3): CombinationRow(data, t_value, (0, 3)),
                               (1, 2): CombinationRow(data, t_value, (1, 2)),
                               (1, 3): CombinationRow(data, t_value, (1, 3)),
                               (2, 3): CombinationRow(data, t_value, (2, 3))}
        self.assertEqual(matrix.total_uncovered, excepted_uncovered, "Total uncovered number is wrong.")
        self.assertEqual(matrix.total_covered_more_than_ones, excepted_covered_more_than_ones,
                         "Total uovered_more_than_ones number is wrong.")
        self.assertEqual(len(matrix.hash_table), excepted_row_size, "Matrix has wrong row size")
        self.assertEqual(len(matrix.uncovered_rows), excepted_row_size, "Matrix has wrong uncovered row size")
        for key in matrix.hash_table:
            with self.subTest(combination=key):
                self.assertTrue(combination_row_equals(matrix.hash_table[key], excepted_hash_table[key]))


class CombinationMatrixTest(unittest.TestCase):

    def setUp(self):
        self.data = [3, 3, 3, 4]
        self.t_value = 2
        self.matrix = CombinationMatrix(self.data, self.t_value)
        self.excepted_hash_table = {(0, 1): CombinationRow(self.data, self.t_value, (0, 1)),
                                    (0, 2): CombinationRow(self.data, self.t_value, (0, 2)),
                                    (0, 3): CombinationRow(self.data, self.t_value, (0, 3)),
                                    (1, 2): CombinationRow(self.data, self.t_value, (1, 2)),
                                    (1, 3): CombinationRow(self.data, self.t_value, (1, 3)),
                                    (2, 3): CombinationRow(self.data, self.t_value, (2, 3))}

    def test_cover_solution_row(self):
        solution_row = [1, 0, 2, 3]
        excepted_uncovered = 57
        excepted_covered_more_than_ones = 0
        excepted_uncovered_row_size = 6
        self.excepted_hash_table[0, 1].cover_cell((1, 0))
        self.excepted_hash_table[0, 2].cover_cell((1, 2))
        self.excepted_hash_table[0, 3].cover_cell((1, 3))
        self.excepted_hash_table[1, 2].cover_cell((0, 2))
        self.excepted_hash_table[1, 3].cover_cell((0, 3))
        self.excepted_hash_table[2, 3].cover_cell((2, 3))
        self.matrix.cover_solution_row(solution_row)
        self.assertEqual(excepted_uncovered, self.matrix.total_uncovered, "Total uncovered number is wrong.")
        self.assertEqual(excepted_covered_more_than_ones, self.matrix.total_covered_more_than_ones,
                         "Total uovered_more_than_ones number is wrong.")
        self.assertEqual(excepted_uncovered_row_size, len(self.matrix.uncovered_rows),
                         "Matrix has wrong uncovered row size")
        for key in self.matrix.hash_table:
            with self.subTest(combination=key):
                self.assertTrue(combination_row_equals(self.matrix.hash_table[key], self.excepted_hash_table[key]))
        solution_row = [0, 0, 2, 3]
        self.matrix.cover_solution_row(solution_row)
        solution_row = [0, 1, 2, 3]
        self.matrix.cover_solution_row(solution_row)
        solution_row = [0, 2, 2, 3]
        self.matrix.cover_solution_row(solution_row)
        solution_row = [1, 1, 2, 3]
        self.matrix.cover_solution_row(solution_row)
        solution_row = [1, 2, 2, 3]
        self.matrix.cover_solution_row(solution_row)
        solution_row = [2, 0, 2, 3]
        self.matrix.cover_solution_row(solution_row)
        solution_row = [2, 1, 2, 3]
        self.matrix.cover_solution_row(solution_row)
        solution_row = [2, 2, 2, 3]
        self.matrix.cover_solution_row(solution_row)
        excepted_uncovered_row_size = 5
        excepted_uncovered = 41
        excepted_covered_more_than_ones = 13
        self.assertEqual(excepted_uncovered, self.matrix.total_uncovered, "Total uncovered number is wrong.")
        self.assertEqual(excepted_covered_more_than_ones, self.matrix.total_covered_more_than_ones,
                         "Total uovered_more_than_ones number is wrong.")
        self.assertEqual(excepted_uncovered_row_size, len(self.matrix.uncovered_rows),
                         "Matrix has wrong uncovered row size")

    def test_cover_combination(self):
        solution_row = [1, 0, 2, 3]
        self.matrix.cover_combination(solution_row, (1, 0))
        excepted_uncovered_row_size = 6
        excepted_uncovered = 58
        excepted_covered_more_than_ones = 0
        self.excepted_hash_table[0, 1].cover_cell((1, 0))
        self.excepted_hash_table[0, 2].cover_cell((1, 2))
        self.excepted_hash_table[0, 3].cover_cell((1, 3))
        self.excepted_hash_table[1, 2].cover_cell((0, 2))
        self.excepted_hash_table[1, 3].cover_cell((0, 3))
        self.assertEqual(excepted_uncovered, self.matrix.total_uncovered, "Total uncovered number is wrong.")
        self.assertEqual(excepted_covered_more_than_ones, self.matrix.total_covered_more_than_ones,
                         "Total uovered_more_than_ones number is wrong.")
        self.assertEqual(excepted_uncovered_row_size, len(self.matrix.uncovered_rows),
                         "Matrix has wrong uncovered row size")
        for key in self.matrix.hash_table:
            with self.subTest(combination=key):
                self.assertTrue(combination_row_equals(self.matrix.hash_table[key], self.excepted_hash_table[key]))

        self.matrix.cover_combination(solution_row, (1, 0))
        excepted_covered_more_than_ones = 5
        self.assertEqual(excepted_uncovered, self.matrix.total_uncovered, "Total uncovered number is wrong.")
        self.assertEqual(excepted_covered_more_than_ones, self.matrix.total_covered_more_than_ones,
                         "Total uovered_more_than_ones number is wrong.")
        self.assertEqual(excepted_uncovered_row_size, len(self.matrix.uncovered_rows),
                         "Matrix has wrong uncovered row size")

    def test_uncover_solution_row(self):
        solution_row = [1, 0, 2, 3]
        self.matrix.cover_solution_row(solution_row)
        self.matrix.uncover_solution_row(solution_row)
        excepted_uncovered_row_size = 6
        excepted_uncovered = 63
        excepted_covered_more_than_ones = 0
        self.assertEqual(excepted_uncovered, self.matrix.total_uncovered, "Total uncovered number is wrong.")
        self.assertEqual(excepted_covered_more_than_ones, self.matrix.total_covered_more_than_ones,
                         "Total uovered_more_than_ones number is wrong.")
        self.assertEqual(excepted_uncovered_row_size, len(self.matrix.uncovered_rows),
                         "Matrix has wrong uncovered row size")
        for key in self.matrix.hash_table:
            with self.subTest(combination=key):
                self.assertTrue(combination_row_equals(self.matrix.hash_table[key], self.excepted_hash_table[key]))

        self.matrix.cover_solution_row(solution_row)
        self.matrix.cover_solution_row(solution_row)
        self.excepted_hash_table[0, 1].cover_cell((1, 0))
        self.excepted_hash_table[0, 2].cover_cell((1, 2))
        self.excepted_hash_table[0, 3].cover_cell((1, 3))
        self.excepted_hash_table[1, 2].cover_cell((0, 2))
        self.excepted_hash_table[1, 3].cover_cell((0, 3))
        self.excepted_hash_table[2, 3].cover_cell((2, 3))
        self.matrix.uncover_solution_row(solution_row)
        excepted_uncovered_row_size = 6
        excepted_uncovered = 57
        excepted_covered_more_than_ones = 0
        self.assertEqual(excepted_uncovered, self.matrix.total_uncovered, "Total uncovered number is wrong.")
        self.assertEqual(excepted_covered_more_than_ones, self.matrix.total_covered_more_than_ones,
                         "Total uovered_more_than_ones number is wrong.")
        self.assertEqual(excepted_uncovered_row_size, len(self.matrix.uncovered_rows),
                         "Matrix has wrong uncovered row size")
        for key in self.matrix.hash_table:
            with self.subTest(combination=key):
                self.assertTrue(combination_row_equals(self.matrix.hash_table[key], self.excepted_hash_table[key]))

    def test_uncover_combination(self):
        solution_row = [1, 0, 2, 3]
        self.matrix.cover_solution_row(solution_row)
        excepted_uncovered_row_size = 6
        excepted_uncovered = 62
        excepted_covered_more_than_ones = 0
        self.excepted_hash_table[2, 3].cover_cell((2, 3))
        self.matrix.uncover_combination(solution_row, (0, 1))
        self.assertEqual(excepted_uncovered, self.matrix.total_uncovered, "Total uncovered number is wrong.")
        self.assertEqual(excepted_covered_more_than_ones, self.matrix.total_covered_more_than_ones,
                         "Total uovered_more_than_ones number is wrong.")
        self.assertEqual(excepted_uncovered_row_size, len(self.matrix.uncovered_rows),
                         "Matrix has wrong uncovered row size")
        for key in self.matrix.hash_table:
            with self.subTest(combination=key):
                self.assertTrue(combination_row_equals(self.matrix.hash_table[key], self.excepted_hash_table[key]))

    def test_uncover(self):
        solution_row = [1, 0, 2, 3]
        self.matrix.cover_solution_row(solution_row)
        self.matrix.cover_solution_row(solution_row)
        excepted_uncovered_row_size = 6
        excepted_uncovered = 63
        excepted_covered_more_than_ones = 0
        self.matrix.uncover()
        self.assertEqual(excepted_uncovered, self.matrix.total_uncovered, "Total uncovered number is wrong.")
        self.assertEqual(excepted_covered_more_than_ones, self.matrix.total_covered_more_than_ones,
                         "Total uovered_more_than_ones number is wrong.")
        self.assertEqual(excepted_uncovered_row_size, len(self.matrix.uncovered_rows),
                         "Matrix has wrong uncovered row size")
        for key in self.matrix.hash_table:
            with self.subTest(combination=key):
                self.assertTrue(combination_row_equals(self.matrix.hash_table[key], self.excepted_hash_table[key]))



import random
import logging

from avocado.core.output import LOG_UI
from avocado.core.output import Throbber

from .Solver import Solver
from .CombinationMatrix import CombinationMatrix

ITERATIONS_SIZE = 600
LOG = LOG_UI.getChild("Cit")
LOG.setLevel(logging.INFO)


class Cit:

    def __init__(self, input_data, t_value, constraints):
        """
        Creation of CombinationMatrix from user input

        :param input_data: parameters from user
        :param t_value: size of one combination
        :param constraints: constraints of combinations
        """
        self.data = input_data
        self.t_value = t_value
        # CombinationMatrix creation
        self.combination_matrix = CombinationMatrix(input_data, t_value)
        # Creation of solver and simplification of constraints
        self.solver = Solver(input_data, constraints)
        # Combinations which do not match to the constraints are disabled
        self.solver.clean_hash_table(self.combination_matrix, t_value)
        self.final_matrix = []
        self.__throbber = Throbber()

    def final_matrix_init(self):
        """
        Creation of the first solution. This solution is the start of searching
        for the best solution

        :return: solution matrix (list(list))
        """
        self.final_matrix = [self.create_random_row_with_constraints()]
        self.combination_matrix.cover_solution_row(self.final_matrix[0])
        while self.combination_matrix.total_uncovered != 0:
            if self.combination_matrix.total_uncovered < self.combination_matrix.total_covered_more_than_ones:
                new_row = self.compute_row()
            else:
                new_row = self.compute_row_using_hamming_distance()
            self.combination_matrix.cover_solution_row(new_row)
            self.final_matrix.append(new_row)
        return self.final_matrix

    def compute(self):
        """
        Searching for the best solution. It creates one solution and from that,
        it tries to create smaller solution. This searching process is limited
        by ITERATIONS_SIZE. When ITERATIONS_SIZE is 0 the last found solution is
        the best solution.

        :return: The best solution
        """
        self.final_matrix = self.final_matrix_init()
        matrix = [x[:] for x in self.final_matrix]
        iterations = ITERATIONS_SIZE
        step_size = 1
        deleted_rows = []
        while step_size != 0:
            for i in range(step_size):
                delete_row = matrix.pop(random.randint(0, len(matrix) - 1))
                self.combination_matrix.uncover_solution_row(delete_row)
                deleted_rows.append(delete_row)
            LOG.debug("I'm trying solution with size %s and %s iterations",
                      len(matrix), iterations)
            matrix, is_better_solution = self.find_better_solution(iterations, matrix)
            if is_better_solution:
                self.final_matrix = matrix[:]
                deleted_rows = []
                step_size *= 2
                LOG.debug("-----solution with size %s was found-----\n",
                          len(matrix))
                iterations = ITERATIONS_SIZE
            else:
                LOG.debug("-----solution with size %s was not found-----\n",
                          len(matrix))
                for i in range(step_size):
                    self.combination_matrix.cover_solution_row(deleted_rows[i])
                    matrix.append(deleted_rows[i])
                if step_size > 1:
                    step_size = 1
                else:
                    step_size = 0

        return self.final_matrix

    def find_better_solution(self, counter, matrix):
        """
        Changing the matrix to cover all combinations

        :param counter: maximum number of changes in the matrix
        :param matrix: matrix to be changed
        :return: new matrix and is changes have been successful?
        """
        while self.combination_matrix.total_uncovered != 0:
            LOG.debug(self.__throbber.render(), extra={"skip_newline": True})
            solution, row_index, _ = self.use_random_algorithm(matrix)
            if len(solution) != 0:
                self.combination_matrix.uncover_solution_row(matrix[row_index])
                self.combination_matrix.cover_solution_row(solution)
                matrix[row_index] = solution
            if counter == 0:
                return matrix, False
            counter -= 1
        return matrix, True

    def use_random_algorithm(self, matrix):
        """
        Applies one of these algorithms to the matrix.
        It chooses algorithm by random in proportion 1:1:8

        :param matrix: matrix to be changed
        :return: new row of matrix, index of row inside matrix and parameters which has been changed
        """
        switch = random.randint(0, 9)
        if switch == 0:
            solution, row_index, parameters = self.change_one_value(matrix)
        elif switch == 1:
            solution, row_index, parameters = self.change_one_column(matrix)
        else:
            solution, row_index, parameters = self.cover_missing_combination(matrix)
        return solution, row_index, parameters

    def compute_row(self):
        """
        Computation of one row which covers most of combinations

        :return: new solution row
        """
        is_valid_row = False
        while not is_valid_row:
            possible_parameters = list(self.combination_matrix.uncovered_rows)
            row = [-1] * len(self.data)
            while len(possible_parameters) != 0:
                # finding uncovered combination
                combination_parameters_index = random.randint(0, len(possible_parameters) - 1)
                combination_parameters = possible_parameters[combination_parameters_index]
                del possible_parameters[combination_parameters_index]
                combination_row = self.combination_matrix.get_row(combination_parameters)
                possible_combinations = list(combination_row.get_all_uncovered_combinations())
                combination_index = random.randint(0, len(possible_combinations) - 1)
                combination = possible_combinations[combination_index]
                is_parameter_used = False
                # Are parameters already used in row?
                for i in combination_parameters:
                    if row[i] != -1:
                        is_parameter_used = True
                        break
                if is_parameter_used:
                    continue
                row_copy = row.copy()
                # Is combination matches the constraints?
                for index, parameter in enumerate(combination_parameters):
                    row_copy[parameter] = combination[index]
                if self.combination_matrix.is_valid_solution(row_copy):
                    row = row_copy
            # Filling in of free spaces inside the row
            for index, r in enumerate(row):
                if r == -1:
                    is_valid = False
                    while not is_valid:
                        row[index] = random.randint(0, self.data[index] - 1)
                        is_valid = self.combination_matrix.is_valid_solution(row)
            is_valid_row = self.combination_matrix.is_valid_solution(row)

        return row

    def cover_missing_combination(self, matrix):
        """
        Randomly finds one missing combination. This combination puts into each
        row of the matrix. The row with the best coverage is the solution

        :param matrix: matrix to be changed
        :return: solution, index of solution inside matrix and parameters which has been changed
        """
        parameters, combination = self.get_missing_combination_random()
        best_uncover = float("inf")
        best_solution = []
        best_row_index = 0
        for row_index in range(len(matrix)):
            solution = [x for x in matrix[row_index]]
            for index, item in enumerate(parameters):
                solution[item] = combination[index]
            if self.combination_matrix.is_valid_combination(solution, parameters):
                self.combination_matrix.uncover_combination(matrix[row_index], parameters)
                self.combination_matrix.cover_combination(solution, parameters)
                if self.combination_matrix.total_uncovered < best_uncover:
                    best_uncover = self.combination_matrix.total_uncovered
                    best_solution = solution
                    best_row_index = row_index
                self.combination_matrix.uncover_combination(solution, parameters)
                self.combination_matrix.cover_combination(matrix[row_index], parameters)
                if best_uncover == 0:
                    break
        return best_solution, best_row_index, parameters

    def get_missing_combination_random(self):
        """
        Randomly finds one missing combination.

        :return: parameter of combination and values of combination
        """
        possible_parameters = list(self.combination_matrix.uncovered_rows)
        combination_parameters_index = random.randint(0, len(possible_parameters) - 1)
        combination_parameters = possible_parameters[combination_parameters_index]
        combination_row = self.combination_matrix.get_row(combination_parameters)
        possible_combinations = list(combination_row.get_all_uncovered_combinations())
        combination_index = random.randint(0, len(possible_combinations) - 1)
        combination = possible_combinations[combination_index]
        return combination_parameters, combination

    def change_one_column(self, matrix):
        """
        Randomly choose one column of the matrix. In each cell of this column
        changes value. The row with the best coverage is the solution.

        :param matrix: matrix to be changed
        :return: solution, index of solution inside matrix and parameters which has been changed
        """
        column_index = random.randint(0, len(self.data) - 1)
        best_uncover = float("inf")
        best_solution = []
        best_row_index = 0
        for row_index in range(len(matrix)):
            try:
                solution, row_index, parameters = self.change_one_value(matrix, row_index, column_index)
            except ValueError:
                continue
            self.combination_matrix.uncover_combination(matrix[row_index], parameters)
            self.combination_matrix.cover_combination(solution, parameters)
            if self.combination_matrix.total_uncovered < best_uncover:
                best_uncover = self.combination_matrix.total_uncovered
                best_solution = solution
                best_row_index = row_index
            self.combination_matrix.uncover_combination(solution, parameters)
            self.combination_matrix.cover_combination(matrix[row_index], parameters)
            if best_uncover == 0:
                break
        return best_solution, best_row_index, [column_index]

    def change_one_value(self, matrix, row_index=None, column_index=None):
        """
        Change one cell inside the matrix

        :param matrix: matrix to be changed
        :param row_index: row inside matrix. If it's None it is chosen randomly
        :param column_index: column inside matrix. If it's None it is chosen randomly
        :return: solution, index of solution inside matrix and parameters which has been changed
        """
        is_cell_chosen = True
        if row_index is None:
            is_cell_chosen = False
            row_index = random.randint(0, len(matrix) - 1)
        row = [x for x in matrix[row_index]]
        if column_index is None:
            is_cell_chosen = False
            column_index = random.randint(0, len(row) - 1)
        possible_numbers = list(range(0, row[column_index])) + list(
            range(row[column_index] + 1, self.data[column_index]))
        row[column_index] = random.choice(possible_numbers)
        while not self.combination_matrix.is_valid_combination(row, [column_index]):
            possible_numbers.remove(row[column_index])
            if len(possible_numbers) == 0:
                if is_cell_chosen:
                    raise ValueError("Selected cell can't be changed")
                column_index = random.randint(0, len(row) - 1)
                row_index = random.randint(0, len(matrix) - 1)
                row = [x for x in matrix[row_index]]
                possible_numbers = list(range(0, row[column_index])) + list(
                    range(row[column_index] + 1, self.data[column_index]))
            row[column_index] = random.choice(possible_numbers)
        return row, row_index, [column_index]

    def compute_row_using_hamming_distance(self):
        """
        :return: row with the biggest hamming distance from final matrix
        """
        row_1 = self.create_random_row_with_constraints()
        row_2 = self.create_random_row_with_constraints()
        if self.compute_hamming_distance(row_1) >= self.compute_hamming_distance(row_2):
            return row_1
        else:
            return row_2

    def compute_hamming_distance(self, row):
        """
        :return: hamming distance of row from final matrix
        """
        distance = 0
        for final_row in self.final_matrix:
            for index, cell in enumerate(final_row):
                if row[index] != cell:
                    distance += 1
        return distance

    def create_random_row_with_constraints(self):
        """
        Create a new test-case random row, and the row meets the constraints.

        :return: new random row
        :rtype: list
        """
        data_size = len(self.data)
        row = [-1]*data_size

        for parameter in random.sample(range(data_size), data_size):
            possible_values = self.solver.get_possible_values(row, parameter)
            value_choice = random.choice(possible_values)
            row[parameter] = value_choice
        return row



import re


class Parser:

    @staticmethod
    def parse(file_object):
        """
        Parsing of input file with parameters and constraints

        :param file_object: input file for parsing
        :return: array of parameters and set of constraints
        """

        rx_parameter_name = re.compile(r"(.+?)\[")
        rx_parameter_value = re.compile(r"\[(.+?)\]")
        rx_constraint = re.compile(r".+?!=.+?")
        parameters = []
        constraints = set()
        is_parameters = None

        for line in file_object:
            line = line.strip()
            if not line:
                continue
            if line == "PARAMETERS":
                is_parameters = True
                continue
            if line == "CONSTRAINTS":
                is_parameters = False
                continue
            if is_parameters is None:
                raise ValueError("Invalid file")
            if is_parameters:
                # Searching for parameter name
                match = rx_parameter_name.search(line)
                if match is None:
                    raise ValueError('Parameter name is missing')
                parameter_name = match.group(1).strip()

                # Searching for parameter_values
                match = rx_parameter_value.search(line)
                if match is None:
                    raise ValueError('Parameter values are missing')
                parameter_values = [x.strip() for x in match.group(1).split(',')]
                if len(parameter_values) != len(set(parameter_values)):
                    raise ValueError('Parameter values has duplicities')

                parameters.append((parameter_name, parameter_values))
            else:
                constraints_data = line.split("||")
                array = []
                for constraint in constraints_data:
                    value_index = 1
                    if not rx_constraint.match(constraint):
                        raise ValueError("Invalid format of constraint")
                    constraint_data = [x.strip() for x in constraint.split("!=")]
                    # Searching for parameter name in constraint
                    constraint_name = next((i for i, p in enumerate(parameters)
                                            if p[0] == constraint_data[0]), None)
                    if constraint_name is None:
                        # If first value inside constraint wasn't parameter name try second value
                        try:
                            constraint_name = next((i for i, p in
                                                    enumerate(parameters)
                                                    if p[0] == constraint_data[1]))
                            value_index = 0
                        except StopIteration:
                            raise ValueError('Name in constraint not match with names in parameters')
                    try:
                        # Checking if value in constraint is matching with parameter values
                        constraint_value = parameters[constraint_name][1]\
                            .index(constraint_data[value_index])
                    except ValueError:
                        raise ValueError('Value in constraint not match with values in parameters')
                    array.append((constraint_name, constraint_value))
                array.sort(key=lambda x: int(x[0]))
                constraints.add(tuple(array))
        return parameters, constraints



import unittest

from avocado_varianter_cit.Solver import Solver


class SolverTest(unittest.TestCase):

    """
    Test for compute_constraints function
    """

    def test_compute_constraints_without_secret_constraint(self):
        """
        Test that, function shouldn't change constraints if there isn't any secret constraint

        """
        parameters = [3, 3, 3]
        constraints = {((0, 0), (2, 0)), ((0, 1), (1, 1)), ((1, 0), (2, 0))}
        solver = Solver([], [])
        solver.data = parameters
        solver.constraints = constraints
        solver.read_constraints()
        solver.compute_constraints()
        self.assertEqual(solver.constraints, constraints, "compute_constraints change constraints without secret "
                                                          "constraint")

    def test_compute_constraints_new_constraint_with_more_than_one_value_for_one_parameter(self):
        """
        Test that, function shouldn't change constraints. It founds new constraint,
        but the constraint has one parameter with two values. This means that it isn't secreted constraint.
        """
        parameters = [3, 3, 3]
        constraints = {((0, 0), (1, 0)), ((0, 1), (1, 1)), ((1, 2), (2, 0))}
        solver = Solver([], [])
        solver.data = parameters
        solver.constraints = constraints
        solver.read_constraints()
        solver.compute_constraints()
        self.assertEqual(solver.constraints, constraints, "compute_constraints change constraints without secret "
                                                          "constraint")

    def test_compute_constraints_detect_invalid_constraints(self):
        """
        Test that, function should raise an exception, if there is invalid_constraints
        """
        parameters = [3, 3, 3]
        constraints = {((0, 0),), ((0, 1),), ((0, 2),)}
        solver = Solver([], [])
        solver.data = parameters
        solver.constraints = constraints
        solver.read_constraints()
        self.assertRaises(ValueError, solver.compute_constraints)

    def test_compute_constraints_with_one_secrete_constraint(self):
        """
        Test that, function should find new constraint
        """
        parameters = [3, 3, 3, 3]
        constraints = {((0, 0), (2, 0)), ((0, 1), (1, 1), (2, 0)), ((0, 2), (3, 2))}
        solver = Solver([], [])
        solver.data = parameters
        solver.constraints = constraints
        solver.read_constraints()
        solver.compute_constraints()
        expectation = {((0, 0), (2, 0)), ((0, 1), (1, 1), (2, 0)), ((0, 2), (3, 2)),
                       ((1, 1), (2, 0), (3, 2))}
        self.assertEqual(solver.constraints, expectation, "compute_constraints didn't find secret constraints")

    # Test for simplify_constraints function

    def test_simplify_constraints_constraints_without_simplification(self):
        """
        Test that, function do not delete important constraints
        """
        parameters = [3, 3, 3, 3]
        constraints = {((0, 0), (2, 0)), ((0, 1), (1, 1), (2, 0)), ((0, 2), (3, 2))}
        solver = Solver([], [])
        solver.data = parameters
        solver.constraints = constraints
        solver.read_constraints()
        solver.simplify_constraints()
        self.assertEqual(solver.constraints, constraints, "simplify_constraints deleted some important constraints")

    def test_simplify_constraints_constraints_with_simplification(self):
        """
        Test that, function do not delete important constraints
        """
        parameters = [3, 3, 3, 3]
        constraints = {((0, 0), (2, 0)), ((0, 1), (2, 0), (1, 1)),
                       ((0, 2), (2, 0)), ((2, 0),)}
        solver = Solver([], [])
        solver.data = parameters
        solver.constraints = constraints
        solver.read_constraints()
        solver.simplify_constraints()
        expectation = {((2, 0),)}
        self.assertEqual(solver.constraints, expectation, "simplify_constraints deleted some important constraints")

    # Test of Minimum forbidden tuple algorithm

    def test_solver_without_secrete_constraints(self):
        """
        Test that, solver didn't change constraints if there isn't any secret constraint
        """
        parameters = [3, 3, 3]
        constraints = {((0, 0), (2, 0)), ((0, 1), (1, 1)), ((1, 0), (2, 0))}
        solver = Solver(parameters, constraints)
        self.assertEqual(solver.constraints, constraints, "solver change constraints without secret "
                                                          "constraint")

    def test_solver_constraints_without_simplification(self):
        """
        Test that, solver do not delete important constraints
        """
        parameters = [3, 3, 3, 3]
        constraints = {((0, 0), (2, 0)), ((0, 1), (1, 1), (2, 0)), ((0, 2), (3, 2))}
        solver = Solver(parameters, constraints)
        self.assertEqual(solver.constraints, constraints, "solver deleted some important constraints")

    def test_solver_constraints_with_simplification_and_secrete(self):
        """
        Test that, Minimum forbidden tuple algorithm can find and simplify constraints
        """
        parameters = [3, 3, 3, 3]
        constraints = {((0, 0), (1, 0)), ((0, 0), (1, 2)), ((0, 1), (3, 0)),
                       ((0, 2), (3, 0)), ((1, 1), (3, 0))}
        solver = Solver(parameters, constraints)
        expectation = {((0, 0), (1, 0)), ((0, 0), (1, 2)), ((3, 0),)}
        self.assertEqual(solver.constraints, expectation, "solver can not compute and simplify constraints")


if __name__ == '__main__':
    unittest.main()



# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
# Authors: Amador Pahim <amador@pahim.org>
#          Bestoun S. Ahmed <bestoon82@gmail.com>
#          Cleber Rosa <crosa@redhat.com>

import logging
import os
import sys

from avocado.core import exit_codes, varianter
from avocado.core.future.settings import settings
from avocado.core.output import LOG_UI
from avocado.core.plugin_interfaces import CLI, Varianter
from avocado.core.tree import TreeNode
from avocado_varianter_cit.Cit import LOG, Cit
from avocado_varianter_cit.Parser import Parser

#: The default order of combinations
DEFAULT_ORDER_OF_COMBINATIONS = 2


class VarianterCitCLI(CLI):

    """
    CIT Varianter options
    """

    name = 'cit'
    description = "CIT Varianter options for the 'run' subcommand"

    def configure(self, parser):

        for name in ("run", "variants"):
            subparser = parser.subcommands.choices.get(name, None)
            if subparser is None:
                continue
            subparser.add_argument_group('CIT varianter options')
            settings.register_option(section="{}.cit".format(name),
                                     key='parameter_file',
                                     metavar='PATH',
                                     help_msg='Paths to a parameter file',
                                     parser=subparser,
                                     default=None,
                                     long_arg='--cit-parameter-file')

            help_msg = "Order of combinations. Maximum number is 6"
            settings.register_option(section="{}.cit".format(name),
                                     key='combination_order',
                                     key_type=int,
                                     parser=subparser,
                                     help_msg=help_msg,
                                     metavar='ORDER',
                                     default=DEFAULT_ORDER_OF_COMBINATIONS,
                                     long_arg='--cit-order-of-combinations')

    def run(self, config):
        if config.get('variants.debug'):
            LOG.setLevel(logging.DEBUG)


class VarianterCit(Varianter):

    """
    Processes the parameters file into variants
    """

    name = 'cit'
    description = "CIT Varianter"

    def initialize(self, config):
        subcommand = config.get('subcommand')
        self.variants = None  # pylint: disable=W0201
        order = config.get("{}.cit.combination_order".format(subcommand))
        if order and order > 6:
            LOG_UI.error("The order of combinations is bigger then 6")
            self.error_exit(config)

        section_key = "{}.cit.parameter_file".format(subcommand)
        cit_parameter_file = config.get(section_key)
        if cit_parameter_file is None:
            return
        else:
            cit_parameter_file = os.path.expanduser(cit_parameter_file)
            if not os.access(cit_parameter_file, os.R_OK):
                LOG_UI.error("parameter file '%s' could not be found or "
                             "is not readable", cit_parameter_file)
                self.error_exit(config)

        try:
            parameters, constraints = Parser.parse(open(cit_parameter_file))
        except ValueError as details:
            LOG_UI.error("Cannot parse parameter file: %s", details)
            self.error_exit(config)

        input_data = [len(parameter[1]) for parameter in parameters]

        cit = Cit(input_data, order, constraints)
        final_list = cit.compute()
        self.headers = [parameter[0] for parameter in parameters]  # pylint: disable=W0201
        results = [[parameters[j][1][final_list[i][j]] for j in range(len(final_list[i]))]
                   for i in range(len(final_list))]
        self.variants = []  # pylint: disable=W0201
        for combination in results:
            self.variants.append(dict(zip(self.headers, combination)))

    @staticmethod
    def error_exit(config):
        if config.get('subcommand') == 'run':
            sys.exit(exit_codes.AVOCADO_JOB_FAIL)
        else:
            sys.exit(exit_codes.AVOCADO_FAIL)

    def __iter__(self):
        if self.variants is None:
            return

        variant_ids = []
        for variant in self.variants:
            variant_ids.append("-".join([variant.get(key)
                                         for key in self.headers]))

        for vid, variant in zip(variant_ids, self.variants):
            yield {"variant_id": vid,
                   "variant": [TreeNode('', variant)],
                   "paths": ['/']}

    def __len__(self):
        return sum(1 for _ in self.variants) if self.variants else 0

    def update_defaults(self, defaults):
        pass

    def to_str(self, summary, variants, **kwargs):
        """
        Return human readable representation

        The summary/variants accepts verbosity where 0 means silent and
        maximum is up to the plugin.

        :param summary: How verbose summary to output (int)
        :param variants: How verbose list of variants to output (int)
        :param kwargs: Other free-form arguments
        :rtype: str
        """
        if not self.variants:
            return ""
        out = []

        if variants:
            # variants == 0 means disable, but in plugin it's brief
            out.append("CIT Variants (%s):" % len(self))
            for variant in self:
                out.extend(varianter.variant_to_str(variant, variants - 1,
                                                    kwargs, False))
        return "\n".join(out)



import random
import unittest
from copy import copy

from avocado_varianter_cit.Cit import Cit
from avocado_varianter_cit.CombinationMatrix import CombinationMatrix
from avocado_varianter_cit.Solver import Solver


class CitInitialization(unittest.TestCase):

    def test_initialization(self):
        parameters = [3, 3, 3, 3]
        constraints = {((0, 0), (2, 0)), ((0, 1), (1, 1), (2, 0)), ((0, 2), (3, 2))}
        t_value = 2
        solver = Solver(parameters, constraints)
        combination_matrix = CombinationMatrix(parameters, t_value)
        solver.clean_hash_table(combination_matrix, t_value)
        cit = Cit(parameters, t_value, constraints)
        self.assertEqual(combination_matrix, cit.combination_matrix, "The initialization of cit algorithm is wrong")


class CitTests(unittest.TestCase):

    def setUp(self):
        parameters = [3, 3, 3, 3]
        constraints = {((0, 0), (2, 0)), ((0, 1), (1, 1), (2, 0)), ((0, 2), (3, 2))}
        t_value = 2
        self.cit = Cit(parameters, t_value, constraints)

    def test_create_random_row_with_constraints(self):
        for _ in range(0, 10):
            row = self.cit.create_random_row_with_constraints()
            with self.subTest(random_row=row):
                self.assertTrue(self.cit.combination_matrix.is_valid_solution(row), "New random row is not valid")

    def test_compute_hamming_distance(self):
        self.cit.final_matrix.append([1, 0, 1, 2])
        self.cit.final_matrix.append([2, 1, 1, 0])
        row = [2, 0, 3, 2]
        expected_distance = 5
        self.assertEqual(expected_distance, self.cit.compute_hamming_distance(row), "Wrong hamming distance")

    def test_final_matrix_init(self):
        combination_matrix = copy(self.cit.combination_matrix)
        final_matrix = self.cit.final_matrix_init()

        expected_total_uncovered = 0
        expected_uncovered_rows = {}
        self.assertEqual(expected_total_uncovered, self.cit.combination_matrix.total_uncovered,
                         "Final matrix don't cover all combinations")
        self.assertEqual(expected_uncovered_rows, self.cit.combination_matrix.uncovered_rows,
                         "Final matrix don't cover all combination rows")

        for row in final_matrix:
            combination_matrix.cover_solution_row(row)
        self.assertEqual(expected_total_uncovered, self.cit.combination_matrix.total_uncovered,
                         "Final matrix don't cover all combinations but CIT thinks it does")
        self.assertEqual(expected_uncovered_rows, self.cit.combination_matrix.uncovered_rows,
                         "Final matrix don't cover all combination rows but CIT thinks it does")

    def test_change_one_value_random(self):
        final_matrix = self.cit.final_matrix_init()
        row, row_index, column_index = self.cit.change_one_value(final_matrix)
        self.assertNotEqual(final_matrix[row_index][column_index[0]], row[column_index[0]], "Value did not change")
        row[column_index[0]] = final_matrix[row_index][column_index[0]]
        self.assertEqual(final_matrix[row_index], row, "Different value was changed")

    def test_change_one_value_with_index(self):
        final_matrix = self.cit.final_matrix_init()
        expected_row_index = 2
        expected_column_index = 0
        function_state = True
        row, row_index, column_index = (None, None, None)
        try:
            row, row_index, column_index = self.cit.change_one_value(final_matrix, row_index=expected_row_index,
                                                                     column_index=expected_column_index)
        except ValueError:
            function_state = False
        if function_state:
            self.assertEqual(expected_column_index, column_index[0], "Column index is wrong")
            self.assertEqual(expected_row_index, row_index, "Row index is wrong")
            self.assertNotEqual(final_matrix[row_index][column_index[0]], row[column_index[0]], "Value did not change")
            row[column_index[0]] = final_matrix[row_index][column_index[0]]
            self.assertEqual(final_matrix[row_index], row, "Different value was changed")
        else:
            self.assertIsNone(row)
            self.assertIsNone(row_index)
            self.assertIsNone(column_index)

    def test_change_one_column(self):
        final_matrix = self.cit.final_matrix_init()
        while self.cit.combination_matrix.total_uncovered == 0:
            delete_row = final_matrix.pop(random.randint(0, len(final_matrix) - 1))
            self.cit.combination_matrix.uncover_solution_row(delete_row)
        expected_total_covered_more_than_ones = self.cit.combination_matrix.total_covered_more_than_ones
        expected_total_uncovered = self.cit.combination_matrix.total_uncovered
        expected_uncovered_rows = copy(self.cit.combination_matrix.uncovered_rows)
        row, row_index, column_index = self.cit.change_one_column(final_matrix)
        self.assertEqual(expected_total_uncovered, self.cit.combination_matrix.total_uncovered, "Coverage was change")
        self.assertEqual(expected_total_covered_more_than_ones,
                         self.cit.combination_matrix.total_covered_more_than_ones,
                         "Coverage was change")
        self.assertEqual(expected_uncovered_rows, self.cit.combination_matrix.uncovered_rows, "Coverage was change")
        self.assertNotEqual(final_matrix[row_index][column_index[0]], row[column_index[0]], "Value did not change")
        row[column_index[0]] = final_matrix[row_index][column_index[0]]
        self.assertEqual(final_matrix[row_index], row, "Different value was changed")

    def test_get_missing_combination_random(self):
        final_matrix = self.cit.final_matrix_init()
        while self.cit.combination_matrix.total_uncovered == 0:
            delete_row = final_matrix.pop(random.randint(0, len(final_matrix) - 1))
            self.cit.combination_matrix.uncover_solution_row(delete_row)
        combination_parameters, combination = self.cit.get_missing_combination_random()
        self.assertEqual(0, self.cit.combination_matrix.hash_table[combination_parameters].hash_table[combination],
                         "Combination is already covered")

    def test_cover_missing_combination(self):
        final_matrix = self.cit.final_matrix_init()
        while self.cit.combination_matrix.total_uncovered == 0:
            delete_row = final_matrix.pop(random.randint(0, len(final_matrix) - 1))
            self.cit.combination_matrix.uncover_solution_row(delete_row)
        expected_total_covered_more_than_ones = self.cit.combination_matrix.total_covered_more_than_ones
        expected_total_uncovered = self.cit.combination_matrix.total_uncovered
        expected_uncovered_rows = copy(self.cit.combination_matrix.uncovered_rows)
        row, row_index, parameters = self.cit.cover_missing_combination(final_matrix)
        self.assertEqual(expected_total_uncovered, self.cit.combination_matrix.total_uncovered, "Coverage was change")
        self.assertEqual(expected_total_covered_more_than_ones,
                         self.cit.combination_matrix.total_covered_more_than_ones,
                         "Coverage was change")
        self.assertEqual(expected_uncovered_rows, self.cit.combination_matrix.uncovered_rows, "Coverage was change")
        self.assertTrue(final_matrix[row_index][parameters[0]] != row[parameters[0]] or
                        final_matrix[row_index][parameters[1]] != row[parameters[1]], "Value did not change")
        row[parameters[0]] = final_matrix[row_index][parameters[0]]
        row[parameters[1]] = final_matrix[row_index][parameters[1]]
        self.assertEqual(final_matrix[row_index], row, "Different value was changed")



import unittest
from avocado_varianter_cit.CombinationRow import CombinationRow


class RowInitialization(unittest.TestCase):

    def test_combination_row_initialization(self):
        """
        Test of proper initialization
        """
        data = [3, 3, 3, 4]
        parameters = (1, 3)
        t_value = 2
        row = CombinationRow(data, t_value, parameters)
        excepted_uncovered = 12
        excepted_covered_more_than_ones = 0
        excepted_hash_table = {(0, 0): 0, (0, 1): 0, (0, 2): 0, (0, 3): 0,
                               (1, 0): 0, (1, 1): 0, (1, 2): 0, (1, 3): 0,
                               (2, 0): 0, (2, 1): 0, (2, 2): 0, (2, 3): 0}
        self.assertEqual(row.uncovered, excepted_uncovered, "Uncovered number is wrong.")
        self.assertEqual(row.covered_more_than_ones, excepted_covered_more_than_ones,
                         "Covered_more_than_ones number is wrong.")
        self.assertEqual(row.hash_table, excepted_hash_table, "Hash table is wrong.")


class CombinationRowTest(unittest.TestCase):

    def setUp(self):
        self.data = [3, 3, 3, 4]
        self.parameters = (1, 3)
        self.t_value = 2
        self.row = CombinationRow(self.data, self.t_value, self.parameters)

    # Tests of cover_cell function

    def test_cover_cell_uncovered_value(self):
        self.assertEqual(self.row.cover_cell((0, 0)), (-1, 0), "cover_cell return wrong values")
        self.assertEqual(self.row.uncovered, 11, "cover_cell create wrong uncovered value")
        self.assertEqual(self.row.covered_more_than_ones, 0, "cover_cell create wrong covered_more_than_ones value")
        self.assertEqual(self.row.hash_table[(0, 0)], 1, "cover_cell don't cover values")

    def test_cover_cell_covered_value(self):
        self.row.hash_table[(0, 0)] = 1
        self.row.uncovered = 11
        self.assertEqual(self.row.cover_cell((0, 0)), (0, 1), "cover_cell return wrong values")
        self.assertEqual(self.row.uncovered, 11, "cover_cell create wrong uncovered value")
        self.assertEqual(self.row.covered_more_than_ones, 1, "cover_cell create wrong covered_more_than_ones value")
        self.assertEqual(self.row.hash_table[(0, 0)], 2, "cover_cell don't cover values")

    def test_cover_cell_cover_disabled_value(self):
        self.row.hash_table[(0, 0)] = None
        self.assertEqual(self.row.cover_cell((0, 0)), (0, 0), "cover_cell return wrong values")
        self.assertEqual(self.row.uncovered, 12, "cover_cell create wrong uncovered value")
        self.assertEqual(self.row.covered_more_than_ones, 0, "cover_cell create wrong covered_more_than_ones value")
        self.assertEqual(self.row.hash_table[(0, 0)], None, "cover_cell change disabled value")

    # Tests of uncover_cell function

    def test_uncover_cell_uncovered_value(self):
        self.assertEqual(self.row.uncover_cell((0, 0)), (0, 0), "uncover_cell return wrong values")
        self.assertEqual(self.row.uncovered, 12, "uncover_cell create wrong uncovered value")
        self.assertEqual(self.row.covered_more_than_ones, 0, "uncover_cell create wrong covered_more_than_ones value")
        self.assertEqual(self.row.hash_table[(0, 0)], 0, "uncover_cell change uncovered value")

    def test_uncover_cell_covered_value(self):
        self.row.hash_table[(0, 0)] = 1
        self.row.uncovered = 11
        self.assertEqual(self.row.uncover_cell((0, 0)), (1, 0), "uncover_cell return wrong values")
        self.assertEqual(self.row.uncovered, 12, "uncover_cell create wrong uncovered value")
        self.assertEqual(self.row.covered_more_than_ones, 0, "uncover_cell create wrong covered_more_than_ones value")
        self.assertEqual(self.row.hash_table[(0, 0)], 0, "uncover_cell change uncovered value")

    def test_uncover_cell_covered_more_than_one_value(self):
        self.row.hash_table[(0, 0)] = 2
        self.row.uncovered = 11
        self.row.covered_more_than_ones = 1
        self.assertEqual(self.row.uncover_cell((0, 0)), (0, -1), "uncover_cell return wrong values")
        self.assertEqual(self.row.uncovered, 11, "uncover_cell create wrong uncovered value")
        self.assertEqual(self.row.covered_more_than_ones, 0, "uncover_cell create wrong covered_more_than_ones value")
        self.assertEqual(self.row.hash_table[(0, 0)], 1, "uncover_cell change uncovered value")

    def test_uncover_cell_disabled_value(self):
        self.row.hash_table[(0, 0)] = None
        self.assertEqual(self.row.uncover_cell((0, 0)), (0, 0), "uncover_cell return wrong values")
        self.assertEqual(self.row.uncovered, 12, "uncover_cell create wrong uncovered value")
        self.assertEqual(self.row.covered_more_than_ones, 0, "uncover_cell create wrong covered_more_than_ones value")
        self.assertEqual(self.row.hash_table[(0, 0)], None, "uncover_cell change disabled value")

    # Test of completely_uncover function

    def test_completely_uncover(self):
        self.row.hash_table[(0, 0)] = 1
        self.row.hash_table[(0, 1)] = 2
        self.row.hash_table[(0, 2)] = None
        self.row.completely_uncover()
        self.assertEqual(self.row.uncovered, 11, "completely_uncover create wrong uncovered value")
        self.assertEqual(self.row.covered_more_than_ones, 0,
                         "completely_uncover create wrong covered_more_than_ones value")
        self.assertEqual(self.row.hash_table[(0, 0)], 0, "completely_uncover don't uncover value")
        self.assertEqual(self.row.hash_table[(0, 1)], 0, "completely_uncover don't uncover value")
        self.assertEqual(self.row.hash_table[(0, 2)], None, "completely_uncover change disabled value")

    # Tests of del_cell function

    def test_del_cell_uncovered_value(self):
        self.assertEqual(self.row.del_cell((0, 0)), -1, "del_cell return wrong values")
        self.assertEqual(self.row.uncovered, 11, "del_cell create wrong uncovered value")
        self.assertEqual(self.row.covered_more_than_ones, 0, "del_cell create wrong covered_more_than_ones value")
        self.assertEqual(self.row.hash_table[(0, 0)], None, "del_cell don't disable value")

    def test_del_cell_disabled_value(self):
        self.row.hash_table[(0, 0)] = None
        self.row.uncovered = 11
        self.assertEqual(self.row.del_cell((0, 0)), 0, "del_cell return wrong values")
        self.assertEqual(self.row.uncovered, 11, "del_cell create wrong uncovered value")
        self.assertEqual(self.row.covered_more_than_ones, 0, "del_cell create wrong covered_more_than_ones value")
        self.assertEqual(self.row.hash_table[(0, 0)], None, "del_cell don't disable value")

    # Tests of is_valid function

    def test_is_valid_valid(self):
        self.assertEqual(self.row.is_valid((0, 0)), True, "is_valid return wrong values")

    def test_is_valid_invalid(self):
        self.row.hash_table[(0, 0)] = None
        self.assertEqual(self.row.is_valid((0, 0)), False, "is_valid return wrong values")

    # Test of get_all_uncovered_combinations function

    def test_get_all_uncovered_combinations(self):
        self.row.hash_table[(0, 0)] = None
        self.row.hash_table[(0, 1)] = 1
        self.row.hash_table[(0, 2)] = 2
        self.row.hash_table[(0, 3)] = 3
        ex = [(1, 0), (1, 1), (1, 2), (1, 3),
              (2, 0), (2, 1), (2, 2), (2, 3)]
        self.assertEqual(len(set(self.row.get_all_uncovered_combinations()).intersection(ex)), len(ex))


if __name__ == '__main__':
    unittest.main()



import os
import tempfile
import unittest

from avocado.utils import process

from selftests import AVOCADO, BASEDIR, temp_dir_prefix


class Basic(unittest.TestCase):

    def test_max_variants(self):
        os.chdir(BASEDIR)
        params_path = os.path.join(BASEDIR, 'examples',
                                   'varianter_cit', 'test_params.cit')
        cmd_line = (
            '{0} variants --cit-order-of-combinations=2 '
            '--cit-parameter-file {1}'
        ).format(AVOCADO, params_path)
        result = process.run(cmd_line)
        lines = result.stdout.splitlines()
        self.assertEqual(b'CIT Variants (9):', lines[0])
        self.assertEqual(10, len(lines))
        for i in range(1, len(lines)):
            with self.subTest(combination=lines[i]):
                self.assertIn(b"green", lines[i])


class Run(unittest.TestCase):

    def setUp(self):
        prefix = temp_dir_prefix(__name__, self, 'setUp')
        self.tmpdir = tempfile.TemporaryDirectory(prefix=prefix)

    def test(self):
        os.chdir(BASEDIR)
        params_path = os.path.join(BASEDIR, 'examples',
                                   'varianter_cit', 'test_params.cit')
        test_path = os.path.join(BASEDIR, 'examples',
                                 'tests', 'cit_parameters.py')
        cmd_line = (
            '{0} --show=test run --sysinfo=off --job-results-dir={1} '
            '--cit-order-of-combinations=1 '
            '--cit-parameter-file={2} '
            '-- {3}'
        ).format(AVOCADO, self.tmpdir.name, params_path, test_path)
        result = process.run(cmd_line)
        # all values should be looked for at least once
        self.assertIn(b"PARAMS (key=color, path=*, default=None) => 'green'",
                      result.stdout)
        # all values of shape should be looked for at least once
        self.assertIn(b"PARAMS (key=shape, path=*, default=None) => 'square'",
                      result.stdout)
        self.assertIn(b"PARAMS (key=shape, path=*, default=None) => 'triangle'",
                      result.stdout)
        self.assertIn(b"PARAMS (key=shape, path=*, default=None) => 'circle'",
                      result.stdout)
        # all values of state should be looked for at least once
        self.assertIn(b"PARAMS (key=state, path=*, default=None) => 'liquid'",
                      result.stdout)
        self.assertIn(b"PARAMS (key=state, path=*, default=None) => 'solid'",
                      result.stdout)
        self.assertIn(b"PARAMS (key=state, path=*, default=None) => 'gas'",
                      result.stdout)
        # all values of material should be looked for at least once
        self.assertIn(b"PARAMS (key=material, path=*, default=None) => 'leather'",
                      result.stdout)
        self.assertIn(b"PARAMS (key=material, path=*, default=None) => 'plastic'",
                      result.stdout)
        self.assertIn(b"PARAMS (key=material, path=*, default=None) => 'aluminum'",
                      result.stdout)
        # all values of coating should be looked for at least once
        self.assertIn(b"PARAMS (key=coating, path=*, default=None) => 'anodic'",
                      result.stdout)
        self.assertIn(b"PARAMS (key=coating, path=*, default=None) => 'cathodic'",
                      result.stdout)

    def tearDown(self):
        self.tmpdir.cleanup()


if __name__ == '__main__':
    unittest.main()



import itertools

from .CombinationRow import CombinationRow as Row


class CombinationMatrix:
    """
    CombinationMatrix object stores Rows of combinations into dictionary.
    And also stores which rows are not covered. Keys in dictionary are parameters
    of combinations and values are CombinationRow objects. CombinationMatrix object
    has information about how many combinations are uncovered and how many of them
    are covered more than ones.
    """

    def __init__(self, input_data, t_value):
        """
        :param input_data: list of data from user
        :param t_value: t number from user
        """
        self.hash_table = {}
        self.uncovered_rows = {}
        self.total_uncovered = 0
        self.total_covered_more_than_ones = 0
        # Creation of rows
        for c in itertools.combinations(range(len(input_data)), t_value):
            row = Row(input_data, t_value, c)
            self.total_uncovered += row.uncovered
            self.hash_table[c] = row
            self.uncovered_rows[c] = c

    def cover_solution_row(self, row):
        """
        Cover all combination by one row from possible solution

        :param row: one row from solution
        :return: number of still uncovered combinations
        """
        for key, value in self.hash_table.items():
            val = []
            # Getting combination from solution
            for i in key:
                val.append(row[i])

            uncovered_difference, covered_more_than_ones_difference = value.cover_cell(val)
            # Deleting covered row from uncovered rows
            if value.uncovered == 0:
                self.uncovered_rows.pop(key, None)
            self.total_uncovered += uncovered_difference
            self.total_covered_more_than_ones += covered_more_than_ones_difference

        return self.total_uncovered

    def cover_combination(self, row, parameters):
        """
        Cover combination of specific parameters by one row from possible solution

        :param row: one row from solution
        :param parameters: parameters which has to be covered
        :return: number of still uncovered combinations
        """
        for key, value in self.hash_table.items():
            for parameter in parameters:
                if parameter in key:
                    val = []
                    # Getting combination from solution
                    for i in key:
                        val.append(row[i])

                    uncovered_difference, covered_more_than_ones_difference = value.cover_cell(val)
                    # Deleting covered row from uncovered rows
                    if value.uncovered == 0:
                        self.uncovered_rows.pop(key, None)
                    self.total_uncovered += uncovered_difference
                    self.total_covered_more_than_ones += covered_more_than_ones_difference
                    break
        return self.total_uncovered

    def uncover_solution_row(self, row):
        """
        Uncover all combination by one row from possible solution

        :param row: one row from solution
        :return: number of uncovered combinations
        """
        for key, value in self.hash_table.items():
            val = []
            # Getting combination from solution
            for i in key:
                val.append(row[i])

            uncovered_difference, covered_more_than_ones_difference = value.uncover_cell(val)
            # Adding uncovered row to uncovered rows
            if value.uncovered != 0:
                self.uncovered_rows[key] = key
            self.total_uncovered += uncovered_difference
            self.total_covered_more_than_ones += covered_more_than_ones_difference

        return self.total_uncovered

    def uncover_combination(self, row, parameters):
        """
        Uncover combination of specific parameters by one row from possible solution

        :param row: one row from solution
        :param parameters: parameters which has to be covered
        :return: number of uncovered combinations
        """
        for key, value in self.hash_table.items():
            for parameter in parameters:
                if parameter in key:
                    val = []
                    # Getting combination from solution
                    for i in key:
                        val.append(row[i])

                    uncovered_difference, covered_more_than_ones_difference = value.uncover_cell(val)
                    # Adding uncovered row to uncovered rows
                    if value.uncovered != 0:
                        self.uncovered_rows[key] = key
                    self.total_uncovered += uncovered_difference
                    self.total_covered_more_than_ones += covered_more_than_ones_difference
                    break
        return self.total_uncovered

    def uncover(self):
        """
        Uncover all combinations
        """
        self.total_covered_more_than_ones = 0
        self.total_uncovered = 0
        for _, value in self.hash_table.items():
            value.completely_uncover()
            self.total_uncovered += value.uncovered

    def is_valid_solution(self, row):
        """
        Is the solution row match the constraints.

        :param row: one row from solution
        """
        for key, value in self.hash_table.items():
            val = []
            for i in key:
                val.append(row[i])

            if not value.is_valid(val):
                return False

        return True

    def is_valid_combination(self, row, parameters):
        """
        Is the specific parameters from solution row match the constraints.

        :param row: one row from solution
        :param parameters: parameters from row
        """
        for key, value in self.hash_table.items():
            for parameter in parameters:
                if parameter in key:
                    val = []
                    for i in key:
                        val.append(row[i])

                    if not value.is_valid(val):
                        return False
                    break
        return True

    def del_cell(self, parameters, combination):
        """
        Disable one combination. If combination is disabled it means that
        the combination does not match the constraints

        :param parameters: parameters whose combination is disabled
        :param combination: combination to be disabled
        """
        row = self.hash_table[tuple(parameters)]
        uncovered_difference = row.del_cell(combination)
        if row.uncovered == 0:
            self.uncovered_rows.pop(tuple(parameters), None)
        self.total_uncovered += uncovered_difference

    def get_row(self, key):
        """
        :param key: identifier of row
        :return: CombinationRow
        """
        return self.hash_table[tuple(key)]

    def __eq__(self, other):
        return (self.total_uncovered == other.total_uncovered and
                self.total_covered_more_than_ones == other.total_covered_more_than_ones and
                self.hash_table == other.hash_table)



#!/bin/env python
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
# See LICENSE for more details.
# Authors: Amador Pahim <amador@pahim.org>
#          Bestoun S. Ahmed <bestoon82@gmail.com>
#          Cleber Rosa <crosa@redhat.com>


from setuptools import setup, find_packages


VERSION = open("VERSION", "r").read().strip()

setup(name='avocado-framework-plugin-varianter-cit',
      description='Varianter with combinatorial capabilities',
      version=open("VERSION", "r").read().strip(),
      author='Avocado Developers',
      author_email='avocado-devel@redhat.com',
      url='http://avocado-framework.github.io/',
      packages=find_packages(),
      include_package_data=True,
      install_requires=['avocado-framework==%s' % VERSION],
      test_suite='tests',
      entry_points={
          'avocado.plugins.cli': [
              'varianter_cit = avocado_varianter_cit:VarianterCitCLI',
          ],
          "avocado.plugins.varianter": [
              "varianter_cit = avocado_varianter_cit:VarianterCit",
          ]}
      )
